@@ -984,7 +984,7 @@ static void btrfs_release_delayed_inode(struct btrfs_delayed_node *delayed_node)
 
 	if (delayed_node &&
 	    test_bit(BTRFS_DELAYED_NODE_INODE_DIRTY, &delayed_node->flags)) {
-		ASSERT(delayed_node->root);
+		BUG_ON(!delayed_node->root);
 		clear_bit(BTRFS_DELAYED_NODE_INODE_DIRTY, &delayed_node->flags);
 		delayed_node->count--;
 
@@ -1133,9 +1133,6 @@ __btrfs_commit_inode_delayed_items(struct btrfs_trans_handle *trans,
 	if (ret)
 		return ret;
 
-	ret = btrfs_record_root_in_trans(trans, node->root);
-	if (ret)
-		return ret;
 	ret = btrfs_update_delayed_inode(trans, node->root, path, node);
 	return ret;
 }
@@ -1174,33 +1171,20 @@ static int __btrfs_run_delayed_items(struct btrfs_trans_handle *trans, int nr)
 		ret = __btrfs_commit_inode_delayed_items(trans, path,
 							 curr_node);
 		if (ret) {
+			btrfs_release_delayed_node(curr_node);
+			curr_node = NULL;
 			btrfs_abort_transaction(trans, ret);
 			break;
 		}
 
 		prev_node = curr_node;
 		curr_node = btrfs_next_delayed_node(curr_node);
-		/*
-		 * See the comment below about releasing path before releasing
-		 * node. If the commit of delayed items was successful the path
-		 * should always be released, but in case of an error, it may
-		 * point to locked extent buffers (a leaf at the very least).
-		 */
-		ASSERT(path->nodes[0] == NULL);
 		btrfs_release_delayed_node(prev_node);
 	}
 
-	/*
-	 * Release the path to avoid a potential deadlock and lockdep splat when
-	 * releasing the delayed node, as that requires taking the delayed node's
-	 * mutex. If another task starts running delayed items before we take
-	 * the mutex, it will first lock the mutex and then it may try to lock
-	 * the same btree path (leaf).
-	 */
-	btrfs_free_path(path);
-
 	if (curr_node)
 		btrfs_release_delayed_node(curr_node);
+	btrfs_free_path(path);
 	trans->block_rsv = block_rsv;
 
 	return ret;
